/*
  XVarStar, GCVS catalogue explorer
  Copyright (C) 2001-2008 Ivan Vitjuk 

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>

#include "variable.h"

#define SEARCH_MAGMAX           0
#define SEARCH_MAGMIN           1

struct varstring{
        size_t buflen, cp;
        char *buf;
};

#define HEADER_STRING   "GCVS name    Ra           Dec      Type        "\
                        "Max  Min   Period       Specter\n" \
                        "-------------------------------------------------"\
                        "----------------------------------------\n"

static void my_toupper(char *str)
{
        while(*str){
                *str = toupper(*str);
                str++;
        }
}

static inline int is_variable(struct variable *var)
{
        //return (*(var->exists) != '=' && *(var->exists) != 'N');
        return 1;
}

void varstring_cat(struct varstring *vs, const char *str)
{
        typedef char *(*strcat_t)(char *, char *, size_t);
        size_t len;
        strcat_t str_cat;
        
        if(!vs->buflen)
                str_cat = (strcat_t)strncpy;
        else
                str_cat = (strcat_t)strncat;

        len = strlen(str);
        if(vs->cp+len+1 >= vs->buflen){
                vs->buflen += MAX(1024, len+1);
                vs->buf = (char *)realloc(vs->buf, vs->buflen);
        }
        
        if(vs->buf){
                str_cat(vs->buf, (char *)str, len);
                vs->cp += len+1;
                vs->buf[vs->cp-1] = 0;
        }
}

/*
  adds one star from struct variable *var to one big string
  that will be show by search_callback()
 */
static void var_add_object(struct varstring *vs, struct variable *var)
{
        char buf[256];
        
        /*
          build line
         */
        snprintf(buf, sizeof(buf)-1, 
                 "%10s  %+03d:%02d:%05.2f %+03d:%02d:%02d "\
                 "%10s %5.2f %5.2f %12.6f %s\n",
                 var->gcvs, 
                 var->Ra1950.h, var->Ra1950.m, var->Ra1950.s,
                 var->Dec1950.d, var->Dec1950.m, var->Dec1950.s,
                 var->vartype, var->magmax, var->magmin,
                 var->period, var->sptype);
        buf[sizeof(buf)-1] = '\0';
        
        if(!vs->buf)
                varstring_cat(vs, HEADER_STRING);

        varstring_cat(vs, buf);
}

/*      
  builds one big string which holds list of stars that will
  be shown by search_callback()
 */
static void var_build_string(struct varstring *vs, variable_t *varlist)
{
        struct variable *tmp;

        for(tmp = varlist->head; tmp; tmp = tmp->next)
                if(is_variable(tmp))
                        var_add_object(vs, tmp);
}

/*
  search stars by name
 */
int var_search_name(const char *patt, variable_t *varlist, variable_t *var)
{
        struct variable *tmp;
        char *str;
        
        if(patt && varlist && var){
                for(tmp = varlist->head; tmp; tmp=tmp->next){
                        for(str = tmp->gcvs; *str && *str == ' '; str++);
                        if(strcasecmp(patt, str)==0){
                                var_list_insert(var, tmp);
                                return 1;
                        }
                }
        }
        return 0;
}

/*
  search stars by magnitude
 */
int _var_search_mag(int minmax, float mag, variable_t *varlist, variable_t *var)
{
        struct variable *tmp;        
        float tmpmag;
        int found=0;
        
        if(varlist && var)
                for(tmp = varlist->head; tmp; tmp=tmp->next){
                        tmpmag = (minmax==SEARCH_MAGMIN)?tmp->magmin:tmp->magmax;
                        if(tmpmag <= mag){
                                var_list_insert(var, tmp);
                                found = 1;
                        }
                }
        return found;
}

/*
  search stars by maximal magnutude
 */
int var_search_magmin(float patt, variable_t *varlist, variable_t *var)
{
        return _var_search_mag(SEARCH_MAGMIN, patt, varlist, var);
}

/*
  search stars by minimall magnutude
 */
int var_search_magmax(float patt, variable_t *varlist, variable_t *var)
{
        return _var_search_mag(SEARCH_MAGMAX, patt, varlist, var);
}

/*
  search stars by type
 */
int var_search_type(char *patt, variable_t *varlist, variable_t *var)
{
        struct variable *tmp;
        int found=0;
        char *str;
        
        if(varlist && var && patt && *patt){
                my_toupper(patt);
                for(tmp = varlist->head; tmp; tmp=tmp->next){
                        for(str = tmp->vartype; *str && *str == ' '; str++);
                        if(strstr(str, patt)){
                                var_list_insert(var, tmp);
                                found  = 1;
                        }
                }
        }
        return found;
}

/*
  search stars by constellation
 */
int var_search_const(int patt, variable_t *varlist, variable_t *var)
{
        struct variable *tmp;
        int found=0;

        if(patt && varlist && var)
                for(tmp = varlist->head; tmp; tmp=tmp->next){
                        if(tmp->constell == patt){
                                var_list_insert(var, tmp);
                                found  = 1;
                        }
                }
        return found;
}

/*
  search stars by amplitude
 */
int var_search_amplitude(float patt, variable_t *varlist, variable_t *var)
{
        struct variable *tmp;        
        int found=0;
        
        if(varlist && var)
                for(tmp = varlist->head; tmp; tmp=tmp->next){
                        if(fabs(tmp->magmax-tmp->magmin) >= patt){
                                var_list_insert(var, tmp);
                                found = 1;
                        }
                }
        return found;
}

/*
  find stars by using user specified search criteria
  
  search result is string holding all found stars that must be freed
  
  set of found stars is set product  where each set holds stars
  satisfying one search criteria (name, type, magnitude, amplitude and
  constellation)
 */
char *var_search(variable_t *varlist, 
                 const char *name,
                 const char *type,
                 const char *mag,
                 const char *amp,
                 const char *constell)
{
        int ret = -1, free=0, tmp;
        struct varstring varstring={0, 0, NULL};
        variable_t *var;
        
        if((var = var_list_init())){
                if(name){
                        ret = var_search_name(name, varlist, var);
                        varlist = var;
                        var = var_list_init();
                        free = 1;
                }

                if(ret && type){
                        ret = var_search_type((char *)type, varlist, var);
                        if(free)
                                var_list_free(varlist);
                        varlist = var;
                        var = var_list_init();
                        free = 1;
                }

                if(ret && mag){
                        ret = var_search_magmax(atof(mag), varlist, var);
                        if(free)
                                var_list_free(varlist);
                        varlist = var;
                        var = var_list_init();
                        free = 1;
                }

                if(ret && amp){
                        ret = var_search_amplitude(atof(amp), varlist, var);
                        if(free)
                                var_list_free(varlist);
                        varlist = var;
                        var = var_list_init();
                        free = 1;
                }
                
                if(ret && constell){
                        if(!(tmp = const_get_id_by_shrt(constell))){
                                var_set_error(8);
                                var_error_reason("constellation: %s", constell);
                                ret = 0;
                        }else{
                                ret = var_search_const(tmp, varlist, var);
                                if(free)
                                        var_list_free(varlist);
                                varlist = var;
                                free = 1;
                        }
                }
        }

        if(ret < 0){
                var_set_error(11);
                return NULL;
        }else if(ret == 0){
                var_set_error(9);
                return NULL;
        }else{
                var_build_string(&varstring, varlist);
                if(free)
                        var_list_free(varlist);
        }
                        
        return varstring.buf;
}
